#ifndef LIBATARI800_SOUND_H_
#define LIBATARI800_SOUND_H_

#include <stdio.h>


extern UBYTE *LIBATARI800_Sound_array;

extern unsigned int sound_array_fill;

extern unsigned int sound_hw_buffer_size;

extern double sample_residual;

#endif /* LIBATARI800_SOUND_H_ */
