#ifndef LIBATARI800_INIT_H_
#define LIBATARI800_INIT_H_

int LIBATARI800_Initialise(void);
void LIBATARI800_Exit(void);
int LIBATARI800_ReadConfig(char *option, char *parameters);

#endif /* LIBATARI800_INIT_H_ */
