#! /bin/sh

rm -rf autom4te.cache

aclocal -I m4
autoheader
autoconf
automake --add-missing

rm -rf autom4te.cache

echo
echo "Now you need to run the configure script.  The configure script may take the"
echo "\"--target=TARGET\" option and various \"--enable-FEATURE\" and \"--with-PACKAGE\""
echo "options."
echo
echo "Run \"./configure --help\" to see all available options."
echo "Run \"./configure --help=short\" just to see the Atari800 FEATURE options."
echo "Run \"./configure --target=help\" without a parameter to see the valid targets."
echo
